import urllib
import urllib.parse
from waddle import ParamBunch

from .odbc import OdbcConnector


class SqlServer(OdbcConnector):
    # https://docs.microsoft.com/en-us/sql/connect/odbc/dsn-connection-string-attribute?view=sql-server-ver16
    default_port = 1433
    engine = 'mssql+pyodbc'
    jdbc_class = 'com.microsoft.sqlserver.jdbc.SQLServerDriver'
    default_driver = 'ODBC Driver 17 for SQL Server'
    ping_query = 'select getdate()'

    @classmethod
    def connect(cls, conf: ParamBunch, *args,
                user=None, password=None, host=None, port=None,
                name=None, driver=None, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        user = user or conf.get('user')
        password = password or conf.get('password')
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        name = name or conf.get('name')
        driver = driver or conf.get('driver', cls.default_driver)
        app = cls.value('application_name', kwargs, conf)
        return super().connect(
            conf, driver=driver, server=f'{host},{port}',
            database=name, uid=user, pwd=password,
            trustservercertificate='yes',
            app=app,
        )

    @classmethod
    def sql_alchemy_connection_string(
            cls, conf: ParamBunch, driver=None, *args, **kwargs):
        """
        the sql alchemy connection string used to connect to the database
        """
        driver = driver or conf.get('driver', cls.default_driver)
        app = cls.value('application_name', kwargs, conf)
        return super().sql_alchemy_connection_string(
            conf, driver=driver, trustservercertificate='yes', app=app,
            **kwargs
        )

    @classmethod
    def jdbc_connection_string(cls, conf: ParamBunch, *args, **kwargs):
        """
        the jdbc connection string used to connect to the database
        """
        port = conf.get('port', cls.default_port)
        params = dict(
            databaseName=f'{conf.name}',
            trustServerCertificate='yes',
            **kwargs,
        )
        params = urllib.parse.urlencode(params)
        return f'jdbc:sqlserver://{conf.host}:{port}?{params}'



import urllib
import urllib.parse
from waddle import ParamBunch

from .odbc import OdbcConnector


class Db2i(OdbcConnector):
    engine = 'ibmi'
    jdbc_class = 'com.ibm.as400.access.AS400JDBCDriver'
    default_driver = 'IBM i Access ODBC Driver 64-bit'
    ping_query = 'select current date from sysibm.sysdummy1'
    default_port = 448
    default_trimchar = 1
    default_ssl = 1

    @classmethod
    def connect(cls, conf: ParamBunch, *args, 
                user=None, password=None, host=None,
                driver=None, port=None, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        driver = driver or conf.get('driver', cls.default_driver)
        port = port or conf.get('port', 448)
        host = host or conf.get('host')
        user = user or conf.get('user')
        password = password or conf.get('password')
        cls.ensure('ssl', kwargs, conf)
        cls.ensure('trimchar', kwargs, conf)
        return super().connect(
            conf, driver=driver, system=host, port=f'{port}',
            uid=user, pwd=password, **kwargs
        )

    @classmethod
    def sql_alchemy_connection_string(
            cls, conf: ParamBunch, *args, host=None, **kwargs):
        """
        the sql alchemy connection string used to connect to the database
        """
        host = host or conf.host
        trim_char = cls.value('trimchar', kwargs, conf)
        cls.ensure('ssl', kwargs, conf)
        return super().sql_alchemy_connection_string(
            conf, suffix=f'{host}/', trim_char_fields=trim_char, **kwargs
        )

    @classmethod
    def jdbc_connection_string(cls, conf: ParamBunch, *args, **kwargs):
        """
        the jdbc connection string used to connect to the database
        """
        return f'jdbc:as400://{conf.host}'



import urllib.parse

from waddle import ParamBunch

from .jdbc import JdbcConnector


class Hp3000(JdbcConnector):
    jdbc_class = 'com.minisoft.jdbc.MSJDBCDriver'
    default_port = 32233

    @classmethod
    def connect(cls, conf: ParamBunch, *args, **kwargs):
        jars = kwargs.pop('jars', None) or conf.jars
        return super().connect(conf, jars=jars, **kwargs)

    @classmethod
    def jdbc_connection_string(cls, conf: ParamBunch, *args, **kwargs):
        port = f"{conf.get('port', cls.default_port)}"
        server_type = f"{conf.get('server_type', 0)}"
        lock_retries = f"{conf.get('lock_retries', 10)}"
        schema = {
            f'Schema{i}': f'{x}.SCHEMA.{conf.password}'
            for i, x in enumerate(conf.get('schema') or [])
        }
        params = dict(
            Server=conf.host,
            ServerPort=port,
            ServerType=server_type,
            LockRetries=lock_retries,
            User=conf.user,
            Account=conf.password,
            Group='PUB',
            **schema,
        )
        params = urllib.parse.urlencode(params)
        url = f'jdbc:MSJDBC:///?{params}'
        return url



from waddle import ParamBunch

from .base import BaseConnector


class Mysql(BaseConnector):
    """
    n.b., I don't like the fact that mysql uses `ssl_mode` while
    postgres uses `sslmode`, but thems the breaks (underlying libraries
    are written that way).
    """
    default_port = 3306
    jdbc_class = 'com.mysql.cj.jdbc.Driver'
    engine = 'mysql+mysqldb'
    default_local_infile = 1
    default_sslmode = 'REQUIRED'

    @classmethod
    def connect(cls, conf: ParamBunch, *args,
                user=None, password=None, host=None,
                port=None, name=None, **kwargs):
        import MySQLdb
        user = user or conf.get('user')
        password = password or conf.get('password')
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        name = name or conf.get('name')
        autocommit = cls.as_bool('autocommit', kwargs, conf)
        local_infile = cls.value('local_infile', kwargs, conf)
        connect_timeout = cls.value('connect_timeout', kwargs, conf)
        sslmode = cls.value('sslmode', kwargs, conf)
        return MySQLdb.connect(
            host=host,
            user=user,
            password=password,
            database=name,
            port=port,
            autocommit=autocommit,
            connect_timeout=connect_timeout,
            ssl_mode=sslmode,
            local_infile=local_infile,
            **kwargs,
        )

    @classmethod
    def sql_alchemy_connection_string(
            cls, conf: ParamBunch, *args,
            **kwargs):
        autocommit = cls.as_bool('autocommit', kwargs, conf)
        local_infile = cls.as_bool('local_infile', kwargs, conf)
        connect_timeout = cls.value('connect_timeout', kwargs, conf)
        sslmode = cls.value('sslmode', kwargs, conf)
        return super().sql_alchemy_connection_string(
            conf, *args,
            connect_timeout=connect_timeout,
            ssl_mode=sslmode,
            local_infile=local_infile,
            autocommit=autocommit)
    


from waddle import ParamBunch

from .base import BaseConnector


class JdbcConnector(BaseConnector):
    @classmethod
    def connect(cls, conf: ParamBunch, *args, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        from jaydebeapi import connect
        jars = kwargs.pop('jars', None)
        driver_args = kwargs.pop('driver_args', None)
        st = cls.jdbc_connection_string(conf, *args, **kwargs)
        connection = connect(cls.jdbc_class, st, driver_args, jars)
        return connection




import sys
from waddle import ParamBunch

from .odbc import OdbcConnector


class Netsuite(OdbcConnector):
    default_trust_store = '/opt/netsuite/odbcclient/cert/ca3.cer'
    default_schema = 'NetSuite2.com'
    default_driver = '{netsuite drivers 64bit}'
    default_port = 1708
    engine = 'pyodbc'

    @classmethod
    def connect(cls, conf: ParamBunch, *args,
                user=None, password=None, host=None, port=None,
                name=None, trust_store=None, driver=None, role_id=None,
                **kwargs):
        user = user or conf.get('user')
        password = password or conf.get('password')
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        account_id = host.split('.')[0].replace('-', '_').upper()
        default_trust_store = cls.default_trust_store
        if sys.platform == 'win32':
            default_trust_store = 'system'
        port = f'{port}'
        trust_store = trust_store or conf.trust_store
        trust_store = trust_store or default_trust_store
        name = name or conf.get('name') or cls.default_schema
        driver = driver or conf.get('default_driver') or cls.default_driver
        role_id = role_id or conf.role_id
        return super().connect(
            conf,
            driver=driver,
            host=host,
            port=port,
            truststore=trust_store,
            encrypted=1,
            allowsinglepacketlogout=1,
            sdsn=name,
            uid=user,
            pwd=password,
            customproperties=f'accountid={account_id};roleid={role_id}',
        )




import urllib
import urllib.parse

from waddle import ParamBunch


class BaseConnector:
    """
    connector classes create connections and connection strings
    in the correct format
    """
    engine = None
    default_port = None
    jdbc_class = None
    default_connect_timeout = 10
    default_application_name = 'concentric'
    default_sslmode = 'require'
    default_autocommit = False

    @classmethod
    def as_bool(cls, key, kwargs, conf: ParamBunch):
        value = kwargs.pop(key, None)
        if value is None:
            value = conf.get(key)
        if value is None:
            value = getattr(cls, f'default_{key}', None)
        return value

    @classmethod
    def value(cls, key, kwargs, conf: ParamBunch):
        value = kwargs.pop(key, None)
        value = value or conf.get(key)
        value = value or getattr(cls, f'default_{key}', None)
        return value

    @classmethod
    def ensure(cls, key, kwargs, conf: ParamBunch):
        kwargs.setdefault(
            key,
            conf.get(key, getattr(cls, f'default_{key}', None)))

    @classmethod
    def connect(cls, conf: ParamBunch, *args, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        pass

    @classmethod
    def sql_alchemy_connection_string(cls, conf: ParamBunch, *args,
                                      user=None, password=None,
                                      host=None, name=None,
                                      port=None, suffix=None, **kwargs):
        """
        the sql alchemy connection string used to connect to the database
        """
        user = user or conf.get('user')
        password = password or conf.get('password')
        password = urllib.parse.quote(password)
        host = host or conf.get('host')
        name = name or conf.get('name')
        port = port or conf.get('port', cls.default_port)
        suffix = suffix or f'{host}:{port}/{name}'
        url = f'{cls.engine}://{user}:{password}@{suffix}'
        if kwargs:
            params = urllib.parse.urlencode(kwargs)
            url = f'{url}?{params}'
        return url

    @classmethod
    def jdbc_connection_string(cls, conf: ParamBunch, *args, **kwargs):
        """
        the jdbc connection string used to connect to the database
        """
        pass



from waddle import ParamBunch

from .base import BaseConnector


class Snowflake(BaseConnector):
    engine = 'snowflake'

    @classmethod
    def connect(cls, conf: ParamBunch, *args, user=None, password=None,
                name=None, **kwargs):
        from snowflake.connector import connect
        name = name or conf.name
        return connect(
            user=user or conf.user,
            password=password or conf.password,
            account=name or conf.name)

    @classmethod
    def sql_alchemy_connection_string(
            cls, conf: ParamBunch, *args,
            user=None, password=None, name=None,
            database=None, schema=None, warehouse=None, **kwargs):
        name = name or conf.name
        pieces = []
        database = database or conf.database
        schema = schema or conf.schema
        warehouse = warehouse or conf.warehouse
        if database:
            pieces.append(database)
            if schema:
                pieces.append(schema)
        junk = '/'.join(pieces)
        suffix = f'{name}/{junk}'
        return super().sql_alchemy_connection_string(
            conf, suffix=suffix, warehouse=warehouse, **kwargs)



import ssl
from ssl import SSLContext, PROTOCOL_SSLv23, CERT_OPTIONAL
from waddle import ParamBunch

from .base import BaseConnector


class Vertica(BaseConnector):
    default_port = 5433
    engine = 'vertica+vertica_python'
    jdbc_class = 'com.vertica.jdbc.Driver'
    default_connection_timeout = 30
    default_autocommit = True

    @classmethod
    def connect(cls, conf: ParamBunch, *args,
                user=None, password=None, host=None, port=None, name=None,
                search_path=None, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        import vertica_python as vertica
        # ssl_context = SSLContext(PROTOCOL_SSLv23)
        # ssl_context.verify_mode = CERT_OPTIONAL
        # ssl_context.check_hostname = False
        user = user or conf.user
        password = password or conf.password
        host = host or conf.host
        name = name or conf.name
        port = port or conf.port or cls.default_port
        connect_timeout = cls.value('connect_timeout', kwargs, conf)
        autocommit = cls.as_bool('autocommit', kwargs, conf)
        connection = vertica.connect(
            host=host,
            port=port,
            database=name,
            user=user,
            password=password,
            connection_timeout=connect_timeout,
            autocommit=autocommit)
        search_path = search_path or conf.get('search_path')
        if search_path:
            with connection.cursor() as cursor:
                cursor.execute(f'set search_path = {", ".join(search_path)}')
        return connection

    @classmethod
    def sql_alchemy_connection_string(cls, conf: ParamBunch, *args, **kwargs):
        """
        provides the connection string to connect to sql alchemy
        """
        cls.ensure('connect_timeout', kwargs, conf)
        cls.ensure('autocommit', kwargs, conf)
        st = super().sql_alchemy_connection_string(conf, *args, **kwargs)
        return st

    @classmethod
    def jdbc_connection_string(cls, conf: ParamBunch, *args, **kwargs):
        """
        the jdbc connection string used to connect to the database
        """
        port = conf.get('port', cls.default_port)
        return f'jdbc:vertica://@{conf.host}:{port}/{conf.name}'



from waddle import ParamBunch

from .base import BaseConnector


class Oracle(BaseConnector):
    default_port = 1521
    engine = 'oracle+cx_oracle'
    jdbc_class = 'oracle.jdbc.driver.OracleDriver'

    @classmethod
    def connect(cls, conf: ParamBunch, *args,
                user=None, password=None, host=None, port=None,
                schema=None, sid=None, service_name=None,
                dsn=None, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        from cx_Oracle import makedsn, connect as oci_connect  # pylint: disable=E0611
        user = user or conf.get('user')
        password = password or conf.get('password')
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        service_name = service_name or conf.get('service_name')
        sid = sid or conf.get('sid')
        dsn = dsn or conf.get('dsn')
        if not dsn:
            if service_name and not sid:
                dsn = makedsn(host, port, service_name=service_name)
            else:
                dsn = makedsn(host, port, sid)
        connection = oci_connect(user, password, dsn, **kwargs)
        schema = schema or conf.get('schema')
        if schema:
            with connection.cursor() as cursor:
                cursor.execute(
                    f'alter session set current_schema={schema}'
                    "  nls_date_format='yyyy-mm-dd hh24:mi:ss'"
                    "  nls_timestamp_format='yyyy-mm-dd hh24:mi:ss'"
                )
        else:
            with connection.cursor() as cursor:
                cursor.execute(
                    'alter session set'
                    "  nls_date_format='yyyy-mm-dd hh24:mi:ss'"
                    "  nls_timestamp_format='yyyy-mm-dd hh24:mi:ss'"
                )
        return connection

    @classmethod
    def sql_alchemy_connection_string(
            cls, conf: ParamBunch, *args,
            host=None, port=None, sid=None, service_name=None,
            dsn=None, **kwargs):
        """
        the sql alchemy connection string used to connect to the database
        """
        from cx_Oracle import makedsn
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        service_name = conf.get('service_name')
        sid = sid or conf.get('sid')
        dsn = dsn or conf.get('dsn')
        if not dsn:
            if service_name and not dsn:
                dsn = makedsn(host, port, service_name=service_name)
            else:
                dsn = makedsn(host, port, sid)
        return super().sql_alchemy_connection_string(
            conf, suffix=dsn, **kwargs)

    @classmethod
    def jdbc_connection_string(cls, conf: ParamBunch, *args,
                               host=None, port=None, sid=None,
                               service_name=None, dsn=None, **kwargs):
        """
        the jdbc connection string used to connect to the database
        """
        from cx_Oracle import makedsn
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        service_name = conf.get('service_name')
        sid = sid or conf.get('sid')
        dsn = dsn or conf.get('dsn')
        if service_name and not dsn:
            dsn = makedsn(conf.host, port, service_name=service_name)
        if dsn:
            return f'jdbc:oracle:thin:@{dsn}'
        return f'jdbc:oracle:thin:@{host}:{port}/{sid}'



import atexit
import threading
import logging
from contextlib import contextmanager

from waddle import ParamBunch
from .db2i import Db2i
from .hp3000 import Hp3000
from .mssql import SqlServer
from .mysql import Mysql
from .netsuite import Netsuite
from .oracle import Oracle
from .postgres import Postgres
from .redshift import Redshift
from .snowflake import Snowflake
from .vertica import Vertica


gtl = threading.local()
log = logging.getLogger(__name__)



class ConnectionManager:
    thread_local_key = 'concentric_connection_manager'


    def __init__(self):
        self.conf = ParamBunch()

    def add_config(self, *filenames):
        for filename in filenames:
            self.conf.from_file(filename, True)

    @classmethod
    def get_instance(cls, *filenames):
        global gtl
        manager = getattr(gtl, cls.thread_local_key, None)
        if manager:
            return manager
        manager = cls()
        manager.add_config(*filenames)
        setattr(gtl, cls.thread_local_key, manager)
        return manager

    @classmethod
    def get_connector(cls, alias):
        manager = cls.get_instance()
        conf = manager.conf[alias]
        if conf.engine == 'oracle':
            return conf, Oracle
        if conf.engine == 'db2i':
            return conf, Db2i
        if conf.engine == 'hp3000':
            return conf, Hp3000
        if conf.engine == 'mssql':
            return conf, SqlServer
        if conf.engine == 'mysql':
            return conf, Mysql
        if conf.engine == 'netsuite':
            return conf, Netsuite
        if conf.engine == 'postgres':
            return conf, Postgres
        if conf.engine == 'redshift':
            return conf, Redshift
        if conf.engine == 'snowflake':
            return conf, Snowflake
        if conf.engine == 'vertica':
            return conf, Vertica
        return None, None

    @classmethod
    def connect(cls, alias, ping=False):
        conf, connector = cls.get_connector(alias)
        return connector.connect(conf)


class CachingConnectionManager(ConnectionManager):
    thread_local_key = 'concentric_caching_connection_manager'
    thread_local_cache_key = 'concentric_connection_cache'

    @classmethod
    def test_connection(cls, alias, conn):
        _, engine = cls.get_connector(alias)
        ping_query = getattr(engine, 'ping_query')
        if ping_query:
            try:
                with conn.cursor() as cursor:
                    cursor.execute(ping_query)
                    cursor.fetchall()
            except:
                conn = None
        return conn

    @classmethod
    def connect(cls, alias, use_cache=True, pre_ping=True):
        global gtl
        cache = getattr(gtl, cls.thread_local_cache_key, None) or {}
        if use_cache:
            conn = cache.get(alias)
            if pre_ping and conn:
                if not hasattr(conn, 'ping'):
                    conn = cls.test_connection(alias, conn)
                else:
                    try:
                        conn.ping()
                    except:
                        conn = None
            if conn:
                return conn
        conn = super().connect(alias)
        if use_cache:
            cache[alias] = conn
            setattr(gtl, cls.thread_local_cache_key, cache)
        return conn

    @classmethod
    def cleanup(cls):
        global gtl
        cache = getattr(gtl, cls.thread_local_cache_key, None) or {}
        keys = []
        for alias, conn in cache.items():
            try:
                keys.append(alias)
                conn.close()
            except:
                pass
        for alias in keys:
            cache.pop(alias, None)


@contextmanager
def transactional(session_cls, initial):
    """
    Provide a transactional scope around a series of operations.
    """
    session = session_cls()
    try:
        for x in initial:
            session.execute(x)
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


@contextmanager
def read_only(session_cls, initial, **params):
    """
    Provide a closing scope around a series of operations.
    used for query operations
    """
    session = session_cls(**params)
    try:
        for x in initial:
            session.execute(x)
        yield session
    finally:
        session.close()


class Alchemist:
    """

    Provides thread safe named access to sqlalchemy sessions

    first declare your connection::

        ConnectionMgr.connection("default", "sqlite:///simple.db")

    then connect::

        with ConnectionMgr.session("default") as session:
            session.execute(sql.insert(role, {"name": "user"}))
            session.commit()

    """

    connections = {}
    connections_lock = threading.Lock()

    @classmethod
    def connection(cls, using, force=False, initial=None, **kwargs):
        """
        creates a cached engine and session_cls for db_url
        kwargs are passed on the create_engine
        """
        from sqlalchemy import create_engine
        from sqlalchemy.orm import sessionmaker
        with cls.connections_lock:
            conn = cls.connections.get(using)
            if conn and not force:
                return conn
            conf, connector = ConnectionManager.get_connector(using)
            url = connector.sql_alchemy_connection_string(conf)
            initial = initial or conf.get('initial') or []
            kwargs.setdefault('pool_pre_ping', True)
            engine = create_engine(url, **kwargs)
            session_cls = sessionmaker(bind=engine)
            conn = engine, session_cls, initial
            cls.connections[using] = conn
        return conn

    @classmethod
    def transactional(cls, using, force=False):
        """
        returns a sql alchemy session for the provided
        database connection that should be defined in waddle
        use the ``force`` flag to force this function to return
        a new connection instead of using the cached connection.
        The semantics of a transactional session are such that at the beginning
        of the context, we will begin a transaction, and once the context
        exits, we will either commit or rollback the transaction.
        """
        conn = cls.connection(using, force=force)
        return transactional(conn[1], conn[2])

    @classmethod
    def read_only(cls, name='default', force=False, **params):
        """
        returns a sql alchemy session for the provided
        database connection that should be defined in waddle
        use the ``force`` flag to force this function to return
        a new connection instead of using the cached connection.
        The semantics of a read only session are that no transactional
        management will occur in the session context.
        """
        conn = cls.connection(name, force=force, **params)
        return read_only(conn[1], conn[2])


def cleanup_cached_connections():
    m = CachingConnectionManager.get_instance()
    m.cleanup()


atexit.register(cleanup_cached_connections)



from waddle import ParamBunch

from .base import BaseConnector


class Postgres(BaseConnector):
    # for connection parameters
    # https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS
    default_port = 5432
    engine = 'postgresql'

    @classmethod
    def connect(cls, conf: ParamBunch, *args,
                user=None, password=None, host=None,
                port=None, name=None, connect_timeout=None,
                application_name=None, sslmode=None, **kwargs):
        import psycopg2
        user = user or conf.user
        password = password or conf.password
        host = host or conf.host
        name = name or conf.name
        port = port or conf.port or cls.default_port
        connect_timeout = connect_timeout or conf.get('connect_timeout')
        connect_timeout = connect_timeout or cls.default_connect_timeout
        sslmode = sslmode or conf.get('sslmode') or cls.default_sslmode
        application_name = application_name or conf.get('application_name')
        application_name = application_name or cls.default_application_name
        return psycopg2.connect(
            dbname=name, user=user, password=password,
            host=host, port=port, connect_timeout=connect_timeout,
            sslmode=sslmode, application_name=application_name)

    @classmethod
    def sql_alchemy_connection_string(
            cls, conf: ParamBunch, *args,
            connect_timeout=None,
            application_name=None, sslmode=None,  **kwargs):
        connect_timeout = connect_timeout or conf.get('connect_timeout')
        connect_timeout = connect_timeout or cls.default_connect_timeout
        sslmode = sslmode or conf.get('sslmode') or cls.default_sslmode
        application_name = application_name or conf.get('application_name')
        application_name = application_name or cls.default_application_name
        return super().sql_alchemy_connection_string(
            conf, *args,
            connect_timeout=connect_timeout,
            application_name=application_name,
            sslmode=sslmode)



from waddle import ParamBunch

from .base import BaseConnector


class OdbcConnector(BaseConnector):
    connect_keys = (
        'readonly',
        'autocommit',
        'encoding',
        'ansi',
        'timeout',
    )

    @classmethod
    def connect(cls, conf: ParamBunch, *args, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        from pyodbc import connect
        pieces = []
        params = {}
        for key, value in kwargs.items():
            if key not in cls.connect_keys:
                pieces.append(f'{key}={value}')
            else:
                params[key] = value
        st = ';'.join(pieces)
        connection = connect(st, **params)
        return connection



from waddle import ParamBunch

from .base import BaseConnector


class Redshift(BaseConnector):
    default_port = 5439
    engine = 'redshift+redshift_connector'

    @classmethod
    def connect(cls, conf: ParamBunch, *args,
                user=None, password=None, host=None,
                port=None, name=None, **kwargs):
        import redshift_connector
        user = user or conf.user
        password = password or conf.password
        host = host or conf.host
        name = name or conf.name
        port = port or conf.port or cls.default_port
        return redshift_connector.connect(
            database=name, user=user, password=password,
            host=host, port=port, )

    @classmethod
    def sql_alchemy_connection_string(cls, conf: ParamBunch, *args,
                                      sslmode='require', **kwargs):
        return super().sql_alchemy_connection_string(
            conf, *args, sslmode=sslmode, **kwargs)